/*

DESCRIPTION:
    Main source file for 12VDC->5VDC SMPS non-isolated buck based on full-bridge
    inverter development using PIC16F1769.

SYNTAX:  n/a

KNOWN BUGS:  n/a

COPYRIGHT:  (c) 2017, Brian Cornell, all rights reserved.

DEPENDENCIES:
    main.h - configuration constants for main.c

VERSION:  1.1, production release

MODIFICATION HISTORY:

    28 Oct 17, BWC, created file.
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "main.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Data objects
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
opcf Config;                                //Configuration data



/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts:
 *      + High-side over-current counter
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    if (C2IF)                           //Over-current
    {
        ++Config.ochs;
        C2IF = FALSE;
    }

    return;
}


/*Function Initialize.  Configures pins & peripherals on the MPU.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization

    OSCCON = CFG_OSC_C;                 //Configure system clock
    while (!(HFIOFL));                  //Wait for oscillator to stabilize

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important but miscellaneous settings that must be set at boot and aren't
     * part of a larger group.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPTION_REGbits.nWPUEN = FALSE;      //Individual pin control over weak pull-ups
    Config.sts = STS_BOOT;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set port configuration.  To prevent any aberrant behaviors before 
     * initialization is complete all output latches are set low.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA = 0x00;
    PORTB = 0x00;
    PORTC = 0x00;
    //Port A
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port B
    TRISB = IO_PORTB;
    ANSELB = AD_PORTB;
    WPUB = WP_PORTB;
    ODCONB = OD_PORTB;
    SLRCONB = SR_PORTB;
    INLVLB = IL_PORTB;
    //Port C
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set fault indicator while starting up and confirm inrush bypass is off
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    setInrushbpOFF();
    setIndFAULT();


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Map peripheral pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_COG1_A = PPS_M_COG1_A;            //COG1 channel A (high-side)
    PPS_P_CLC2_Q = PPS_M_CLC2_Q;            //CLC2 is low side
    PPS_P_PWM_3 = PPS_M_PWM_3;              //Gate drive for output switch

#ifdef _USE_PRG2_FE_CLC1
    PPS_P_PRG2_FS = PPS_M_PRG2_FS;          //PRG2 falling source input
    PPS_P_CLC1_Q = PPS_M_CLC1_Q;            //Falling event signal for PRG2
#endif

#ifdef _DEBUG
    PPS_P_COG1_B = PPS_M_COG1_B;            //COG1 channel B (low-side)
#endif


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure FVR & DAC1-4.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    __delay_us(FI_FVR_DLY);

    DAC1CON0 = CFG_DAC_1_C;
    DAC1REFH = CFG_DAC_1_VH;
    DAC1REFL = CFG_DAC_1_VL;
    DAC1LD = TRUE;                      //Load reference value

    DAC2CON0 = CFG_DAC_2_C;
    DAC2REFH = CFG_DAC_2_VH;
    DAC2REFL = CFG_DAC_2_VL;
    DAC2LD = TRUE;                      //Load reference value

    DAC3CON0 = CFG_DAC_3_C;
    DAC3REF = CFG_DAC_3_V;

    DAC4CON0 = CFG_DAC_4_C;
    DAC4REF = CFG_DAC_4_V;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure opamp OPA1 as modulator error amp.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA1CON = CFG_OPA_1_C;
    OPA1NCHS = CFG_OPA_1_N;
    OPA1PCHS = CFG_OPA_1_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure comparator 1 for modulator control loop.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM1CON0 = CFG_CM_1_C0;
    CM1CON1 = CFG_CM_1_C1;
    CM1NSEL = CFG_CM_1_N;
    CM1PSEL = CFG_CM_1_P;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure comparator 2 for over-current sense.  Output used by COG as a
     * falling event input.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM2CON0 = CFG_CM_2_C0;
    CM2CON1 = CFG_CM_2_C1;
    CM2NSEL = CFG_CM_2_N;
    CM2PSEL = CFG_CM_2_P;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure comparator 3 for low-side switch control.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM3CON0 = CFG_CM_3_C0;
    CM3CON1 = CFG_CM_3_C1;
    CM3NSEL = CFG_CM_3_N;
    CM3PSEL = CFG_CM_3_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure ADC baseline operation.  Clock, +/- references, auto-conversion,
     * and justification don't change.  Enable and default connect to DAC4 to
     * calibrate & obtain Vdd values.
     * 
     * Rule:  no assumptions should be made about which analog channel the ADC
     * is connected to.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (ADC_CH_DAC_4 | S_EN_ADC);
    ADCON1 = ADCCFGC1;
    ADCON2 = ADCCFGC2;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer2 to support 10-bit PWMs
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T2CLKCON = CFG_TMR_2_CLK;
    T2CON = CFG_TMR_2_CON;
    T2PR = CFG_TMR_2_PR;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM3 as PWM switch for output gate drive
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM3CON = CFG_PWM_3_CTRL;
    PWM3DCH = (CFG_PWM_3_DCYCLE >> 2);
    PWM3DCL = ((CFG_PWM_3_DCYCLE & 0x03) << 6);
    T2ON = TRUE;                            //Start timer but PWM is left off


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM5 as primary PWM signal
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5CLKCON = CFG_PWM_5_CLOCK;
    PWM5LDCON = CFG_PWM_5_LDATRIG;
    PWM5PH = CFG_PWM_5_PHASE;
    PWM5DC = S_DC_MAX;
    PWM5PR = CFG_PWM_5_FRQ;
    PWM5OF = CFG_PWM_5_OFFSET;
    PWM5TMR = CFG_PWM_5_TMR;
    PWM5CON = CFG_PWM_5_CTRL;               //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM6 as slave PWM signal
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM6CLKCON = CFG_PWM_6_CLOCK;
    PWM6LDCON = CFG_PWM_6_LDATRIG;
    PWM6OFCON = CFG_PWM_6_TRIGGER;
    PWM6PH = CFG_PWM_6_PHASE;
    PWM6DC = CFG_PWM_6_DCYCLE;
    PWM6PR = CFG_PWM_6_FRQ;
    PWM6OF = CFG_PWM_6_OFFSET;
    PWM6TMR = CFG_PWM_6_TMR;
    PWM6CON = CFG_PWM_6_CTRL;               //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure complementary output generator 1 to drive synchronous buck
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1CON0 = CFG_COG_1_CON0;
    COG1CON1 = CFG_COG_1_CON1;
    COG1BLKR = CFG_COG_1_FE_BLANK;
    COG1ASD0 = CFG_COG_1_AS_CONTROL;
    COG1RIS1 = CFG_COG_1_RE_ONTIME;
    COG1FIS0 = (CFG_COG_1_FE_OVRCRNT | CFG_COG_1_FE_MODULATOR);
    COG1FIS1 = CFG_COG_1_FE_MAXDC;
    COG1RSIM1 = CFG_COG_1_RM_ONTIME;
    COG1FSIM0 = (CFG_COG_1_FM_OVRCRNT | CFG_COG_1_FM_MODULATOR);
    COG1FSIM1 = CFG_COG_1_FM_MAXDC;
    COG1DBF = CFG_COG_1_FE_DEADBAND;
    COG1DBR = CFG_COG_1_RE_DEADBAND;
    COG1CON0bits.LD = TRUE;                 //Load count registers


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PRG1 for modulator error-amp compare
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PRG1CON0 = CFG_PRG_1_C0;
    PRG1CON1 = CFG_PRG_1_C1;
    PRG1INS = CFG_PRG_1_VS;
    PRG1CON2 = CFG_PRG_1_C2;
    PRG1RTSS = CFG_PRG_1_RS;
    PRG1FTSS = CFG_PRG_1_FS;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PRG2 for low-side gate drive termination compare
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PRG2CON0 = CFG_PRG_2_C0;
    PRG2CON1 = CFG_PRG_2_C1;
    PRG2INS = CFG_PRG_2_VS;
    PRG2CON2 = CFG_PRG_2_C2;
    PRG2RTSS = CFG_PRG_2_RS;
    PRG2FTSS = CFG_PRG_2_FS;


#ifdef _USE_PRG2_FE_CLC1
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLC1 as FE to PRG2
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC1SEL0 = CFG_CLC_1_DS_0;
    CLC1SEL1 = CFG_CLC_1_DS_1;
    CLC1SEL2 = CFG_CLC_1_DS_2;
    CLC1SEL3 = CFG_CLC_1_DS_3;
    CLC1GLS0 = CFG_CLC_1_LS_0;
    CLC1GLS1 = CFG_CLC_1_LS_1;
    CLC1GLS2 = CFG_CLC_1_LS_2;
    CLC1GLS3 = CFG_CLC_1_LS_3;
    CLC1POL = CFG_CLC_1_POL;
    CLC1CON = CFG_CLC_1_CON;
    CLC1CONbits.EN = TRUE;                  //Enable
#endif

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLC2 as low-side gate drive
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC2SEL0 = CFG_CLC_2_DS_0;
    CLC2SEL1 = CFG_CLC_2_DS_1;
    CLC2SEL2 = CFG_CLC_2_DS_2;
    CLC2SEL3 = CFG_CLC_2_DS_3;
    CLC2GLS0 = CFG_CLC_2_LS_0;
    CLC2GLS1 = CFG_CLC_2_LS_1;
    CLC2GLS2 = CFG_CLC_2_LS_2;
    CLC2GLS3 = CFG_CLC_2_LS_3;
    CLC2POL = CFG_CLC_2_POL;
    CLC2CON = CFG_CLC_2_CON;
    CLC2CONbits.EN = TRUE;                  //Enable


    //Wait for ramps to stabilize and start
    while (!(PRG1CON1bits.RDY & PRG2CON1bits.RDY));
    PRG1CON0bits.GO = TRUE;
    PRG2CON0bits.GO = TRUE;

    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts

    //Enable bridge over-current event capture
    C2IE = TRUE;

    return;
}


/*Function main.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    unsigned int hysBSM, hysOTL, hysOTH, hysOVC;

    //Configure system and pause for inrush & settle time
    Initialize();
    __delay_ms(S_TM_INT);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * An bus over-volt situation could damage the PS when inrush bypass is
     * turned on (if high enough may damage anyway).  If TRUE abort start and 
     * permanently disable PS (e.g. require POR).
     * 
     * Otherwise enable inrush bypass relay and delay > voltage sense filter 
     * circuit time constant so that PS doesn't unnecessarily read a low bus
     * voltage when close to UVLO threshold.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ReadVoltage(PIN_I_DCBUSV, S_RT_BUS, &Config.bus);
    if (Config.bus.vmv > S_LM_BSX) Config.sts = STS_BUS_OVLO;
    else setInrushbpON();
    __delay_ms(S_TM_INRSHBP);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Run-time loop.  Start with DC bus low voltage hysteresis counter at limit
     * so that PS won't attempt a start when initially entering loop if bus
     * voltage is low.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    for (hysBSM = S_LM_BSM_HYS; (Config.sts != STS_BUS_OVLO); )
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Scan operating environment
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReadVoltage(PIN_I_DCBUSV, S_RT_BUS, &Config.bus);
        ReadVoltage(PIN_I_OUTPTV, S_RT_OUT, &Config.out);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Perform limit checks.  Immediate & permanent shutdown if bus OV
         * exists.  Next, always check for bus UV.  Last, perform 'normal' 
         * limit checks only when PS is running.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.bus.vmv > S_LM_BSX)
        {
            Config.sts = STS_BUS_OVLO;
            break;
        }
        if (Config.bus.vmv < S_LM_BSM)              //Low DC bus voltage
        {
            if (++hysBSM > S_LM_BSM_HYS) Config.sts = STS_BUS_UVLO;
        }
        else if (hysBSM) --hysBSM;

        if (Config.sts == STS_RUN)                  //'Normal' limit checks
        {
            while (TRUE)                            //Break on first error
            {
#ifndef _OPEN_LOOP
                if (Config.out.vmv < S_LM_OTL)      //Low output voltage
                {
                    if (++hysOTL > S_LM_OTL_HYS)
                    {
                        Config.sts = STS_OUT_UVLO;
                        break;
                    }
                }
                else if (hysOTL) --hysOTL;

                if (Config.out.vmv > S_LM_OTH)      //High output voltage
                {
                    if (++hysOTH > S_LM_OTH_HYS)
                    {
                        Config.sts = STS_OUT_OVLO;
                        break;
                    }
                }
                else if (hysOTH) --hysOTH;
#endif

                if (Config.ochs > S_LM_OVC)         //High-side over current
                {
                    if (++hysOVC > S_LM_OVC_HYS)
                    {
                        Config.sts = STS_CHL_OCLO;
                        break;
                    }
                }
                else if (hysOVC) --hysOVC;
                Config.ochs = 0;                    //Always clear OC counter

                break;
            }
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                      Shutdown on error
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts > STS_RUN)
        {
            COG1ASD0bits.ASE = TRUE;
            setOutputOFF();
            setIndFAULT();
            __delay_ms(S_TM_STOP);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Attempt restart if bus voltage is above restart value but always
             * attempt restart for non-UVLO codes.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            blinkErrorCode(Config.sts);
            if (Config.sts == STS_BUS_UVLO)
            {
                if (Config.bus.vmv > S_LM_BSR) Config.sts = STS_BOOT;
            }
            else Config.sts = STS_BOOT;
        }

        if (Config.sts == STS_BOOT)
        {
            //Clear counters
            hysBSM = 0;
            hysOTL = 0;
            hysOTH = 0;
            hysOVC = 0;
            Config.ochs = 0;

            PWM5DC = S_DC_MIN;                  //Soft start with min DC
            PWM5LDCONbits.LDA = TRUE;
            COG1ASD0bits.ASE = FALSE;           //Output on
            chgDutyCycle(S_DC_MIN, S_DC_MAX);   //Increase DC limit to max

            //Update status, indicator, and turn on output switch
            setOutputON();
            setIndRUN();
            Config.sts = STS_RUN;
        }
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Run-time loop aborted:  make sure PS is shutdown and require POR to
     * attempt restart.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1ASD0bits.ASE = TRUE;
    setOutputOFF();
    setInrushbpOFF();
    setIndFAULT();
    while (TRUE) blinkErrorCode(Config.sts);
} //End function main


/* Function chgDutyCycle.  Moves the duty cycle from the start to final value 
 * incrementally for soft start.
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void chgDutyCycle(unsigned int start, unsigned int final)
{
    while (TRUE)
    {
        PWM5DC = start;
        PWM5LDCONbits.LDA = TRUE;
        if (start == final) break;

        if (start < final) ++start; else --start;
        __delay_us(FCDC_DCDELAY);
    }

    return;
}


/* Function ReadVoltage.  Reads the value from the analog pin channel, interprets 
 * as a scaled voltage, and stores in the volts object pointed to by rv.
 * 
 * The ADC must be pre-configured to use the FVR x 4 voltage reference (4.096V).
 * 
 * No error checking is performed and only a single sample is taken.  The voltage
 * is stored in 'composite' form for easy integer evaluation of the complete
 * value (e.g. V/mV) by storing the integer value in the 1,000 & 10,000 place
 * of a 16-bit integer and the mV in the 100, 10, and 1.  The discrete volts
 * and mV are also stored in separate integers.
 * 
 * Returns:  none.  The following values are populated in object rv:
 *      vmv=composite value
 *      volt=integer voltage component
 *      mv=millivolts component
 *
 * Parameters:  channel=analog channel.  The pin must be properly configured.
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *
*/
void ReadVoltage(unsigned char channel, unsigned char scale, volts *rv)
{
    static union {
        unsigned char result[SIZ_INT_ADC];
        unsigned int value;
    } adc;

    ADCON0 = (unsigned) (channel | S_EN_ADC);   //Connect to channel & enable ADC
    __delay_us(FRV_ADC_AQT);                    //Refresh sample & hold cap
    GO_nDONE = TRUE;                            //Start conversion
    while (GO_nDONE);                           //Wait for completion
    adc.result[POS_HIG_ADC] = ADRESH;
    adc.result[POS_LOW_ADC] = ADRESL;

    adc.value *= scale;
    rv->vmv = (((adc.value / FRV_ADC_V1V) * 1000) + ((adc.value % FRV_ADC_V1V) * 4));
    rv->volt = rv->vmv / 1000;
    rv->mv = rv->vmv % 1000;

    return;
}


void blinkErrorCode(unsigned char ec)
{
    setIndOFF();
    __delay_ms(S_TM_BLK_SPD);
    for (; (ec); --ec)
    {
        __delay_ms(S_TM_BLK_SPD);
        setIndFAULT();
        __delay_ms(S_TM_BLK_SPD);
        setIndOFF();
    }

    __delay_ms(S_TM_BLK_SPD * 2);
    setIndFAULT();
    return;
}



//End of file